from enum import Enum


class StudyType(Enum):
	BBROWSER=0
	H5_10X=1
	H5AD=2
	MTX_10X=3
	BCS=4
	RDS=5
	TSV=6
	DSP=7
	VISIUM=8
	VIZGEN=9
	COSMX=10
	XENIUM=11
	VISIUM_RDS=12
	VISIUM_ANN=13
	VIZGEN_V2=14
	TILE_DB=15
	PROTEOMICS_QPTIFF=16
	PROTEOMICS_OME_TIF=17


class TechnologyType(Enum):
	SINGLE_CELL='SC'
	LENS_SC='LENS_SC'
	VISIUM='VISIUM'
	DSP='DSP'
	PROTEOMICS='PROTEOMICS'


class Species(Enum):
  HUMAN='human'
  MOUSE='mouse'
  NON_HUMAN_PRIMATE='primate'
  OTHERS='others'


class InputMatrixType(Enum):
  RAW='raw'
  NORMALIZED='normalized'


class StudyUnit(Enum):
	UNIT_RAW='raw'
	UNIT_LOGNORM='lognorm'


INPUT_LENS_SC = {
	StudyType.COSMX.value: {
		'folders': {
			'cell_boundaries': 'celllabels',
			'images': 'rawmorphologyimages'
		},
		'files': {
			'transcripts': 'tx_file.csv',
			'fov_positions': 'fov_positions_file.csv'
		}
	},
	StudyType.VIZGEN.value: {
		'folders': {
			'cell_boundaries': 'cell_boundaries',
			'images': 'images'
		},
		'files': {
			'transcripts': 'detected_transcripts.csv'
		}
	},
	StudyType.XENIUM.value: {
		'folders': {},
		'files': {
			'micron2pixel': 'experiment.xenium',
			'cell_boundaries': 'cell_boundaries.csv.gz',
			'transcripts': 'transcripts.csv.gz',
			'image': 'morphology_mip.ome.tif'
		}
	},
	StudyType.VIZGEN_V2.value: {
		'folders': {
			'images': 'images'
		},
		'files': {
			'transcripts': 'detected_transcripts.csv',
			'cell_boundaries': 'cell_boundaries.parquet'
		}
	},
	StudyType.PROTEOMICS_OME_TIF.value: {
		'folders': {},
		'files': {
			'image': 'image.ome.tif'
		}
	},
	StudyType.PROTEOMICS_QPTIFF.value: {
		'folders': {},
		'files': {
			'image': 'image.qptiff'
		}
	}
}


class ChunkSize(Enum):
  CHUNK_5_MB = 5 * 1024 * 1024
  CHUNK_100_MB = 100 * 1024 * 1024
  CHUNK_500_MB = 500 * 1024 * 1024
  CHUNK_1_GB = 1 * 1024 * 1024 * 1024


"""A python class for performing HTTPS requests (GET, POST)"""

import requests
import base64


class RequestError(Exception):
  """Exception for request error"""
  pass


class HttpsAgent:
  """Wrapper class of requests"""
  def __init__(self, token: str, ssl: bool):
    self.__token = token
    self.__ssl = ssl


  def get(self, url: str, params: dict = None):
    """Perform GET request

      Args:
        url:
          An API endpoint
          Example: https://bioturing.com/api
        params:
          Params of the GET requests, will be encoded to URL's query string
          Example: {"param1": 0, "param2": true}
    """
    if params is None:
      params = {}

    try:
      res = requests.get(
        url=url,
        params=params,
        headers={'bioturing-api-token': self.__token},
        verify=self.__ssl
      )
      return res.json()
    except requests.exceptions.RequestException as e:
      print('Request fail with error: ', e)
      return None


  def post(self, url: str, body: dict = None, check_error=True):
    """
    Perform POST request

    Args:
      url:
        An API endpoint
        Example: https://bioturing.com/api

      body:
        Body of the request
        Example: {"param1": 0, "param2": true}
    """
    if body is None:
      body = {}

    try:
      res = requests.post(
        url=url,
        json=body,
        headers={'bioturing-api-token': self.__token},
        verify=self.__ssl
      )
      res_json = res.json()
      if check_error and res.status_code >= 400:
        if 'traceback' in res_json:
          message = f"Request fail with traceback:\n{res_json['traceback']}"
        else:
          detail = res_json.get('detail', 'Something went wrong, please contact support@bioturing.com.')
          message = f"Request fail with message:\n{detail}"
        raise RequestError(message)
      return res_json
    except requests.exceptions.RequestException as e:
      print('Request fail with error: ', e)
      return None



from . import typing
from . import bbrowserx_connector
from . import lens_bulk_connector
from . import lens_sc_connector
from . import connector

__version__ = '1.7.0'

__all__ = [
  'typing',
  'bbrowserx_connector',
  'lens_bulk_connector',
  'lens_sc_connector',
  'connector',
  '__version__'
]



"""Common function"""

import os
import uuid
import base64
import datetime
import numpy as np


def get_datetime():
  return str(datetime.now()).split('.', maxsplit=1)[0]


def get_uuid():
  return uuid.uuid4().hex


def get_file_size(path):
    return os.path.getsize(path)


def decode_base64_array(base64_str: str, dtype: str, shape=(-1,)):
	decoded = base64.b64decode(base64_str)
	arr = np.frombuffer(decoded, dtype=dtype)
	return arr.reshape(shape)


import pandas as pd
from scipy import sparse

from . import decode_base64_array


def check_result_status(result):
  if not result:
    raise ConnectionError('Connection failed')

  if 'status' not in result:
    raise ValueError(result['detail'])

  if result['status'] != 200:
    if 'message' in result:
      raise Exception(f"Something went wrong: {result['message']}")
    else:
      raise Exception('Something went wrong')


def parse_submission_status(submission_status):
  """
    Parse submission status response
  """
  if not submission_status:
    print('Connection failed')
    return None

  if 'status' not in submission_status:
    if 'detail' in submission_status:
      print (submission_status['detail'])
    else:
      print('Internal server error. Please check server log.')
    return None

  if submission_status['status'] != 200:
    if 'message' in submission_status:
      print(f"Submission failed. {submission_status['message']}")
      return None
    else:
      print('Submission failed.')
      return None

  return submission_status['data']['id']


def parse_query_genes_result(query_genes_result):
  """Parse query genes result
  """
  check_result_status(query_genes_result)

  indptr = decode_base64_array(query_genes_result['data']['indptr'], 'uint64')
  indices = decode_base64_array(query_genes_result['data']['indices'], 'uint32')
  data = decode_base64_array(query_genes_result['data']['data'], 'float32')
  shape = query_genes_result['data']['shape']
  csc_mtx = sparse.csc_matrix((data, indices, indptr), shape=shape)
  return csc_mtx


def dataframe2dictionary(df):
  res = dict()
  res['barcodes'] = df.index.values.tolist()
  for column in df.columns:
    tmp_data = df.loc[:, column].values
    try:
      data = [int(x) for x in tmp_data]
    except:
      data = [str(x) for x in tmp_data]
    res[column] = data
  return res


def read_csv(path, **kwargs):
  df = pd.read_csv(filepath_or_buffer = path, sep='\t', **kwargs)

  if 'index_col' in kwargs:
    if len(df.columns) == 0:
      return pd.read_csv(filepath_or_buffer = path, sep=',', **kwargs)
  else:
    if len(df.columns) < 2:
      return pd.read_csv(filepath_or_buffer = path, sep=',', **kwargs)

  return df


def parse_root_leaf_name(
    ontologies_tree,
    root_name,
    leaf_name,
  ):
  root_ids = []
  for id in ontologies_tree['tree']:
    if ontologies_tree['tree'][id]['name'] == root_name:
      root_ids.append(id)
  for root_id in root_ids:

    children = ontologies_tree['tree'][root_id]['children']
    for child in children:
      if child['name'] == leaf_name:
        leaf_id = child['id']
        return root_id, leaf_id

      grand_children = child['children']
      for grand_child in grand_children:
        if grand_child['name'] == leaf_name:
          leaf_id = grand_child['id']
          return root_id, leaf_id

  raise Exception('Cannot find "{}" - "{}"'.format(root_name, leaf_name))
